@@ -15,18 +15,22 @@
 // You should have received a copy of the GNU General Public License
 // along with ESPMonitor.  If not, see <https://www.gnu.org/licenses/>.
 
+use addr2line::Context;
 use crossterm::{
+    QueueableCommand,
     event::{self, Event, KeyCode, KeyEvent, KeyModifiers},
+    style::{Color, Print, PrintStyledContent, Stylize},
     terminal::{disable_raw_mode, enable_raw_mode},
 };
+use gimli::{read::Reader, EndianRcSlice, RunTimeEndian};
 use lazy_static::lazy_static;
-use serial::{self, BaudRate, SerialPort, SystemPort};
+use object::read::Object;
 use regex::Regex;
+use serial::{self, BaudRate, SerialPort, SystemPort};
 use std::{
-    ffi::{OsString, OsStr},
-    io::{self, ErrorKind, Read, Write},
-    path::Path,
-    process::{Command, Stdio, exit},
+    fs,
+    io::{self, ErrorKind, Read, Write, stdout},
+    process::exit,
     time::{Duration, Instant},
 };
 
@@ -42,8 +46,6 @@ lazy_static! {
         .expect("Failed to parse line separator regex");
     static ref FUNC_ADDR_RE: Regex = Regex::new(r"0x4[0-9a-f]{7}")
         .expect("Failed to parse program address regex");
-    static ref ADDR2LINE_RE: Regex = Regex::new(r"^0x[0-9a-f]+:\s+([^ ]+)\s+at\s+(\?\?|[0-9]+):(\?|[0-9]+)")
-        .expect("Failed to parse addr2line output regex");
 }
 
 macro_rules! rprintln {
@@ -52,11 +54,15 @@ macro_rules! rprintln {
     ($fmt:literal, $($arg:tt)+) => (print!(concat!($fmt, "\r\n"), $($arg)*));
 }
 
-struct SerialState {
+struct Symbols<'a> {
+    obj: object::read::File<'a, &'a [u8]>,
+    context: Context<EndianRcSlice<RunTimeEndian>>,
+}
+
+struct SerialState<'a> {
     unfinished_line: String,
     last_unfinished_line_at: Instant,
-    bin: Option<OsString>,
-    tool_prefix: &'static str,
+    symbols: Option<Symbols<'a>>,
 }
 
 #[cfg(unix)]
@@ -95,7 +101,7 @@ pub fn run(args: AppArgs) -> Result<(), Box<dyn std::error::Error>> {
     result
 }
 
-fn run_child(mut args: AppArgs) -> Result<(), Box<dyn std::error::Error>> {
+fn run_child(args: AppArgs) -> Result<(), Box<dyn std::error::Error>> {
     rprintln!("ESPMonitor {}", env!("CARGO_PKG_VERSION"));
     rprintln!();
     rprintln!("Commands:");
@@ -112,30 +118,45 @@ fn run_child(mut args: AppArgs) -> Result<(), Box<dyn std::error::Error>> {
         settings.set_baud_rate(speed)
     })?;
 
-    if let Some(bin) = args.bin.as_ref() {
-        if Path::new(bin).exists() {
-            rprintln!("Using {} as flash image", bin.to_string_lossy());
-        } else {
-            rprintln!("WARNING: Flash image {} does not exist (you may need to build it)", bin.to_string_lossy());
-        }
-    }
+    let bin_data = args.bin.as_ref().and_then(|bin_name| match fs::read(bin_name) {
+        Ok(bin_data) => {
+            rprintln!("Using {} as flash image", bin_name.to_string_lossy());
+            Some(bin_data)
+        },
+        Err(err) => {
+            rprintln!("WARNING: Unable to open flash image {}: {}", bin_name.to_string_lossy(), err);
+            None
+        },
+    });
+
+    let symbols = bin_data.as_ref().and_then(|bin_data| match load_bin_context(bin_data.as_slice()) {
+        Ok(symbols) => Some(symbols),
+        Err(err) => {
+            rprintln!("WARNING: Failed to parse flash image: {}", err);
+            None
+        },
+    });
 
     if args.reset {
         reset_chip(&mut dev)?;
     }
 
+
     let mut serial_state = SerialState {
         unfinished_line: String::new(),
         last_unfinished_line_at: Instant::now(),
-        bin: args.bin.take(),
-        tool_prefix: args.chip.tool_prefix(),
+        symbols,
     };
 
+    let mut output = stdout();
     let mut buf = [0u8; 1024];
     loop {
         match dev.read(&mut buf) {
-            Ok(bytes) if bytes > 0 => handle_serial(&mut serial_state, &buf[0..bytes])?,
-            Ok(_) => (),
+            Ok(bytes) if bytes > 0 => handle_serial(&mut serial_state, &buf[0..bytes], &mut output)?,
+            Ok(_) => if dev.read_dsr().is_err() {
+                rprintln!("Device disconnected; exiting");
+                break Ok(());
+            },
             Err(err) if err.kind() == ErrorKind::TimedOut => (),
             Err(err) if err.kind() == ErrorKind::WouldBlock => (),
             Err(err) => break Err(err.into()),
@@ -151,6 +172,15 @@ fn run_child(mut args: AppArgs) -> Result<(), Box<dyn std::error::Error>> {
     }
 }
 
+fn load_bin_context(data: &[u8]) -> Result<Symbols, Box<dyn std::error::Error + 'static>> {
+    let obj = object::File::parse(data)?;
+    let context = Context::new(&obj)?;
+    Ok(Symbols {
+        obj,
+        context,
+    })
+}
+
 fn reset_chip(dev: &mut SystemPort) -> io::Result<()> {
     print!("Resetting device... ");
     std::io::stdout().flush()?;
@@ -161,7 +191,7 @@ fn reset_chip(dev: &mut SystemPort) -> io::Result<()> {
     Ok(())
 }
 
-fn handle_serial(state: &mut SerialState, buf: &[u8]) -> io::Result<()> {
+fn handle_serial(state: &mut SerialState, buf: &[u8], output: &mut dyn Write) -> io::Result<()> {
     let data = String::from_utf8_lossy(buf);
     let mut lines = LINE_SEP_RE.split(&data).collect::<Vec<&str>>();
 
@@ -182,8 +212,7 @@ fn handle_serial(state: &mut SerialState, buf: &[u8]) -> io::Result<()> {
             };
 
         if !full_line.is_empty() {
-            let processed_line = process_line(state, full_line);
-            rprintln!("{}", processed_line);
+            output_line(state, full_line, output)?;
             state.unfinished_line.clear();
         }
     }
@@ -192,36 +221,49 @@ fn handle_serial(state: &mut SerialState, buf: &[u8]) -> io::Result<()> {
         state.unfinished_line.push_str(nel);
         state.last_unfinished_line_at = Instant::now();
     } else if !state.unfinished_line.is_empty() && state.last_unfinished_line_at.elapsed() > UNFINISHED_LINE_TIMEOUT {
-        let processed_line = process_line(state, &state.unfinished_line);
-        rprintln!("{}", processed_line);
+        output_line(state, &state.unfinished_line, output)?;
         state.unfinished_line.clear();
     }
 
     Ok(())
 }
 
-fn process_line(state: &SerialState, line: &str) -> String {
-    let mut updated_line = line.to_string();
+fn output_line(state: &SerialState, line: &str, output: &mut dyn Write) -> io::Result<()> {
+    if let Some(symbols) = state.symbols.as_ref() {
+        let mut cur_start_offset = 0;
 
-    if let Some(bin) = state.bin.as_ref() {
         for mat in FUNC_ADDR_RE.find_iter(line) {
-            let cmd = format!("{}addr2line", state.tool_prefix);
-            if let Some(output) = Command::new(&cmd)
-                .args(&[OsStr::new("-pfiaCe"), bin, OsStr::new(mat.as_str())])
-                .stdout(Stdio::piped())
-                .output()
+            let (function, file, lineno) = u64::from_str_radix(&mat.as_str()[2..], 16)
                 .ok()
-                .and_then(|output| String::from_utf8(output.stdout).ok())
-            {
-                if let Some(caps) = ADDR2LINE_RE.captures(&output) {
-                    let name = format!("{} [{}:{}:{}]", mat.as_str().to_string(), caps[1].to_string(), caps[2].to_string(), caps[3].to_string());
-                    updated_line = updated_line.replace(mat.as_str(), &name);
-                }
+                .map(|addr| {
+                    let function = find_function_name(symbols, addr);
+                    let (file, lineno) = find_location(symbols, addr);
+                    (function, file, lineno)
+                })
+                .unwrap_or((None, None, None));
+
+            fn or_qq(s: Option<String>) -> String {
+                s.unwrap_or_else(|| "??".to_string())
             }
+
+            output.queue(Print(line[cur_start_offset..mat.end()].to_string()))?;
+            cur_start_offset = mat.end();
+            let symbolicated_name = format!(" [{}:{}:{}]", or_qq(function), or_qq(file), or_qq(lineno.map(|l| l.to_string())))
+                .with(Color::Yellow);
+            output.queue(PrintStyledContent(symbolicated_name))?;
         }
+
+        if cur_start_offset < line.len() {
+            output.queue(Print(line[cur_start_offset..line.len()].to_string()))?;
+        }
+    } else {
+        output.queue(Print(line.to_string()))?;
     }
 
-    updated_line
+    output.write_all(b"\r\n")?;
+    output.flush()?;
+
+    Ok(())
 }
 
 fn handle_input(dev: &mut SystemPort, key_event: KeyEvent) -> io::Result<()> {
@@ -235,3 +277,23 @@ fn handle_input(dev: &mut SystemPort, key_event: KeyEvent) -> io::Result<()> {
         Ok(())
     }
 }
+
+fn find_function_name(symbols: &Symbols<'_>, addr: u64) -> Option<String> {
+    symbols.context
+        .find_frames(addr)
+        .ok()
+        .and_then(|mut frames| frames.next().ok().flatten())
+        .and_then(|frame| frame.function.and_then(|function| function.name.to_string_lossy().ok().map(|name| name.into_owned())))
+        .or_else(|| symbols.obj.symbol_map().get(addr).map(|sym| sym.name().to_string()))
+}
+
+fn find_location(symbols: &Symbols<'_>, addr: u64) -> (Option<String>, Option<u32>) {
+    symbols.context
+        .find_location(addr)
+        .ok()
+        .map(|location| (
+            location.as_ref().and_then(|location| location.file).map(|file| file.to_string()),
+            location.as_ref().and_then(|location| location.line)
+        ))
+        .unwrap_or((None, None))
+}
